@@ -12,13 +12,18 @@ The following interfaces are currently available:
 """
 # License: Simplified BSD, 2013
 # Author: Jake Vanderplas <jakevdp@cs.washington.edu>
+# See http://github.com/jakevdp/scidb-py for more information
+
+from __future__ import print_function
 
 import abc
-import urllib2
+
+from ._py3k_compat import urlopen, quote, HTTPError, iteritems, string_type
+
 import re
 import csv
 import numpy as np
-from .scidbarray import SciDBArray, SciDBDataShape, ArrayAlias
+from .scidbarray import SciDBArray, SciDBDataShape, ArrayAlias, SDB_IND_TYPE
 from .errors import SHIM_ERROR_DICT
 from .utils import broadcastable
 
@@ -135,6 +140,7 @@ class SciDBInterface(object):
         else:
             # on subsequent calls, increment the array count
             self.array_count += 1
+
         return "{0}{1}_{2:05}".format(arr_key, self.uid, self.array_count)
 
     def _scan_array(self, name, **kwargs):
@@ -217,7 +223,7 @@ class SciDBInterface(object):
         """
         parse = lambda x: ArrayAlias(x) if isinstance(x, SciDBArray) else x
         args = (parse(v) for v in args)
-        kwargs = dict((k, parse(v)) for k, v in kwargs.iteritems())
+        kwargs = dict((k, parse(v)) for k, v in iteritems(kwargs))
         query = query.format(*args, **kwargs)
         return query
 
@@ -230,9 +236,9 @@ class SciDBInterface(object):
 
         Parameters
         ----------
-        query: string
+        query : string
             The query string, with curly-braces to indicate insertions
-        *args, **kwargs:
+        *args, **kwargs :
             Values to be inserted (see below).
 
         Details
@@ -268,7 +274,7 @@ class SciDBInterface(object):
         The dimension/attribute shortcuts are as follows:
 
         - A.d0, A.d1, A.d2... : insert array dimension names.
-          For the above case, '{A.d0}' will be translated to 'i',
+          For the above case, '{A.d0}' will be translated to 'i','
           and '{A.d1}' will be translated to 'j'.
         - A.d0f, A.d1f, A.d2f... : insert fully-qualified dimension names.
           For the above case, '{A.d0f}' will be translated to 'myarray.i',
@@ -327,11 +333,11 @@ class SciDBInterface(object):
 
         Parameters
         ----------
-        shape: tuple or int
+        shape : tuple or int
             The shape of the array
-        dtype: string or list
+        dtype : string or list
             The data type of the array
-        **kwargs:
+        **kwargs :
             Additional keyword arguments are passed to SciDBDataShape.
 
         Returns
@@ -348,11 +354,11 @@ class SciDBInterface(object):
 
         Parameters
         ----------
-        shape: tuple or int
+        shape : tuple or int
             The shape of the array
-        dtype: string or list
+        dtype : string or list
             The data type of the array
-        **kwargs:
+        **kwargs :
             Additional keyword arguments are passed to SciDBDataShape.
 
         Returns
@@ -370,15 +376,15 @@ class SciDBInterface(object):
 
         Parameters
         ----------
-        shape: tuple or int
+        shape : tuple or int
             The shape of the array
-        dtype: string or list
+        dtype : string or list
             The data type of the array
-        lower: float
+        lower : float
             The lower bound of the random sample (default=0)
-        upper: float
+        upper : float
             The upper bound of the random sample (default=1)
-        **kwargs:
+        **kwargs :
             Additional keyword arguments are passed to SciDBDataShape.
 
         Returns
@@ -402,15 +408,15 @@ class SciDBInterface(object):
 
         Parameters
         ----------
-        shape: tuple or int
+        shape : tuple or int
             The shape of the array
-        dtype: string or list
+        dtype : string or list
             The data type of the array
-        lower: float
+        lower : float
             The lower bound of the random sample (default=0)
-        upper: float
+        upper : float
             The upper bound of the random sample (default=2147483647)
-        **kwargs:
+        **kwargs :
             Additional keyword arguments are passed to SciDBDataShape.
 
         Returns
@@ -546,16 +552,16 @@ class SciDBInterface(object):
         ----------
         n : integer
             the number of rows and columns in the matrix
-        dtype: string or list
+        dtype : string or list
             The data type of the array
-        sparse: boolean
+        sparse : boolean
             specify whether to create a sparse array (default=False)
-        **kwargs:
+        **kwargs :
             Additional keyword arguments are passed to SciDBDataShape.
 
         Returns
         -------
-        arr: SciDBArray
+        arr : SciDBArray
             A SciDBArray containint an [n x n] identity matrix
         """
         arr = self.new_array((n, n), dtype, **kwargs)
@@ -659,26 +665,133 @@ class SciDBInterface(object):
                 self.query("store(gesvd({0}, '{1}'),{2})", A, output, ret[-1])
         return tuple(ret)
 
-    def from_array(self, A, **kwargs):
-        """Initialize a scidb array from a numpy array"""
-        # TODO: make this work for other data types
-        if A.dtype != 'double':
-            raise NotImplementedError("from_array only implemented for double")
-        dtype = 'double'
-        data = A.tostring(order='C')
+    def from_array(self, A, instance_id=0, **kwargs):
+        """Initialize a scidb array from a numpy array
+
+        Parameters
+        ----------
+        A : array_like (numpy array or sparse array)
+            input array from which the scidb array will be created
+        instance_id : integer
+            the instance ID used in loading
+            (default=0; see SciDB documentation)
+        **kwargs :
+            Additional keyword arguments are passed to new_array()
+        
+        Returns
+        -------
+        arr : SciDBArray
+            SciDB Array object built from the input array
+        """
+        A = np.asarray(A)
+        instance_id = int(instance_id)
         filename = self._upload_bytes(A.tostring(order='C'))
-        arr = self.new_array(A.shape, 'double', **kwargs)
-        self.query("load({0},'{1}',{2},'{3}')", arr, filename, 0, '(double)')
+        arr = self.new_array(A.shape, A.dtype, **kwargs)
+        self.query("load({0},'{1}',{2},'{3}')",
+                   arr, filename, instance_id,
+                   arr.sdbtype.bytes_fmt)
         return arr
 
-    def toarray(self, A):
+    def from_dataframe(self, A, instance_id=0, **kwargs):
+        """Initialize a scidb array from a pandas dataframe
+
+        Parameters
+        ----------
+        A : pandas dataframe
+            data from which the scidb array will be created.
+        instance_id : integer
+            the instance ID used in loading
+            (default=0; see SciDB documentation)
+        **kwargs :
+            Additional keyword arguments are passed to new_array()
+        
+        Returns
+        -------
+        arr : SciDBArray
+            SciDB Array object built from the input array
+        """
+        # load to SciDB via a record array (with index in 'index' field)
+        A_rec = A.to_records(index=True)
+        A_sdb = self.from_array(A_rec)
+
+        # TODO: rename index if value is passed?
+        if 'dim_names' in kwargs:
+            warnings.warn("from_dataframe: ignoring 'dim_names' argument")
+        kwargs['dim_names'] = [A_rec.dtype.names[0]]
+
+        # redimension the array on the index
+        arr = self.new_array(shape=A_rec.shape,
+                             dtype=A_rec.dtype.descr[1:],
+                             **kwargs)
+        self.query("redimension_store({0},{1})", A_sdb, arr)
+        return arr
+
+    def from_sparse(self, A, instance_id=0, **kwargs):
+        """Initialize a scidb array from a sparse array
+
+        Parameters
+        ----------
+        A : sparse array
+            sparse input array from which the scidb array will be created.
+            Note that this array will internally be converted to COO format.
+        instance_id : integer
+            the instance ID used in loading
+            (default=0; see SciDB documentation)
+        **kwargs :
+            Additional keyword arguments are passed to new_array()
+        
+        Returns
+        -------
+        arr : SciDBArray
+            SciDB Array object built from the input array
+        """
+        try:
+            A = A.tocoo()
+        except:
+            raise ValueError("input must be a scipy.sparse matrix")
+        instance_id = int(instance_id)
+
+        if 'dim_names' not in kwargs:
+            kwargs['dim_names'] = ['i0', 'i1']
+
+        attr_name = 'f0'
+
+        if len(kwargs['dim_names']) != 2:
+            raise ValueError("dim_names must have two dimensions")
+        d1, d2 = kwargs['dim_names']    
+
+        # first flatten the array & indices
+        # We'll treat the general case where A.data can be a record array,
+        # though this will be very uncommon.
+        flat_dtype = [(d1, SDB_IND_TYPE),
+                      (d2, SDB_IND_TYPE)]
+        if A.dtype.names is None:
+            flat_dtype += [('f0', A.dtype)]
+        else:
+            flat_dtype += A.descr
+        M = np.empty(len(A.data), dtype=flat_dtype)
+        M[d1] = A.row
+        M[d2] = A.col
+        M['f0'] = A.data
+        arr_flat = self.from_array(M)
+
+        # redimension the flat array to a sparse array
+        arr = self.new_array(A.shape, A.dtype, **kwargs)
+        self.query("redimension_store({0},{1})", arr_flat, arr)
+        return arr
+
+    def toarray(self, A, transfer_bytes=True):
         """Convert a SciDB array to a numpy array"""
-        return A.toarray()
+        return A.toarray(transfer_bytes=transfer_bytes)
 
     def tosparse(self, A, sparse_fmt='recarray', transfer_bytes=True):
         """Convert a SciDB array to a sparse representation"""
         return A.tosparse(sparse_fmt=sparse_fmt, transfer_bytes=transfer_bytes)
 
+    def todataframe(self, A, transfer_bytes=True):
+        """Convert a SciDB array to a pandas dataframe"""
+        return A.todataframe(transfer_bytes=transfer_bytes)
+
     def from_file(self, filename, **kwargs):
         # TODO: allow creation of arrays from uploaded files
         # TODO: allow creation of arrays from pre-existing files within the
@@ -687,6 +800,7 @@ class SciDBInterface(object):
 
     def _apply_func(self, A, func):
         # TODO: new value name could conflict.  How to generate a unique one?
+        # TODO: add optional ``out`` argument as in numpy
         arr = self.new_array()
         self.query("store(project(apply({A},{func}_{A.a0},{func}({A.a0})),"
                    "{func}_{A.a0}), {arr})",
@@ -762,6 +876,13 @@ class SciDBInterface(object):
     def substitute(self, A, value):
         return A.substitute(value)
 
+    def merge(self, A, B):
+        """Merge two arrays"""
+        # TODO: pre-check for non-matching arrays?
+        arr = self.new_array()
+        self.query("store(merge({0},{1}),{2})", A, B, arr)
+        return arr
+
     def join(self, *args):
         """
         Perform a series of array joins on the arguments
@@ -983,17 +1104,16 @@ class SciDBShimInterface(SciDBInterface):
     def __init__(self, hostname):
         self.hostname = hostname.rstrip('/')
         try:
-            urllib2.urlopen(self.hostname)
-        except urllib2.HTTPError:
+            urlopen(self.hostname)
+        except HTTPError:
             raise ValueError("Invalid hostname: {0}".format(self.hostname))
 
     def _get_uid(self):
         # load a library to get a query id
         session = self._shim_new_session()
-        uid = self._shim_execute_query(session,
-                                       "load_library('dense_linear_algebra')",
-                                       release=True)
-        return uid
+        return self._shim_execute_query(session,
+                                        "load_library('dense_linear_algebra')",
+                                        release=True)
 
     def _execute_query(self, query, response=False, n=0, fmt='auto'):
         # log the query
@@ -1024,13 +1144,13 @@ class SciDBShimInterface(SciDBInterface):
         url = self.hostname + '/' + keyword
         if kwargs:
             url += '?' + '&'.join(['{0}={1}'.format(key, val)
-                                   for key, val in kwargs.iteritems()])
+                                   for key, val in iteritems(kwargs)])
         return url
 
     def _shim_urlopen(self, url):
         try:
-            return urllib2.urlopen(url)
-        except urllib2.HTTPError as e:
+            return urlopen(url)
+        except HTTPError as e:
             Error = SHIM_ERROR_DICT[e.code]
             raise Error("[HTTP {0}] {1}".format(e.code, e.read()))
 
@@ -1048,10 +1168,10 @@ class SciDBShimInterface(SciDBInterface):
     def _shim_execute_query(self, session_id, query, save=None, release=False):
         url = self._shim_url('execute_query',
                              id=session_id,
-                             query=urllib2.quote(query),
+                             query=quote(query),
                              release=int(bool(release)))
         if save is not None:
-            url += "&save={0}".format(urllib2.quote(save))
+            url += "&save={0}".format(quote(save))
 
         # Don't know the accepted way in Python to do something like this.
         # For now, just set the 'debug' attribute on this SciDB array object.
@@ -1060,7 +1180,7 @@ class SciDBShimInterface(SciDBInterface):
 
         result = self._shim_urlopen(url)
         query_id = result.read()
-        return query_id
+        return query_id.decode('UTF-8')
 
     def _shim_cancel(self, session_id):
         url = self._shim_url('cancel', id=session_id)
@@ -1070,6 +1190,9 @@ class SciDBShimInterface(SciDBInterface):
         url = self._shim_url('read_lines', id=session_id, n=n)
         result = self._shim_urlopen(url)
         text_result = result.read()
+        # the following check is for Py3K compatibility
+        if not isinstance(text_result, string_type):
+            text_result = text_result.decode('UTF-8')
         return text_result
 
     def _shim_read_bytes(self, session_id, n):
@@ -1079,7 +1202,7 @@ class SciDBShimInterface(SciDBInterface):
         return bytes_result
 
     def _shim_upload_file(self, session_id, data):
-        # TODO: can this be implemented in urllib2 to remove dependency?
+        # TODO: can this be implemented in urllib to remove dependency?
         import requests
         url = self._shim_url('upload_file', id=session_id)
         result = requests.post(url, files=dict(fileupload=data))
